#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/epoll.h>
#include <time.h>
#include <signal.h>
#include "fproxy.h"

/* TODO support kqueue for FreeBSD */

#define LOGFILE			"/var/log/fproxy.log"
#define ERRFILE			"/var/log/fproxy.err"
#define PIDFILE			"/var/run/fproxy.pid"
#define LISTEN_BACKLOG		1024 /* listen()                           */
#define EPOLL_HINT		1024 /* max # of events from epoll_wait()  */
#define CONN_POOL_SIZE		4096 /* max # of unused conn structs       */
#define EPOLL_TIMEOUT		2000 /* milliseconds                       */
#define PAGE_SIZE		4096 /* chunk size for allocating Strings  */
#define LOOPED_SYSCALL(A)	do { errno = 0; A; } while (errno == EINTR)
#define NULLSTRING		(struct String) { .s = NULL, .l = 0,._l = 0 }

#if 0
#  include <stdio.h>
#  define DEBUGP(fmt,args...)	fprintf(stderr, "[debug]: %s: " fmt "\n", \
                                        __FUNCTION__, ## args)
#else
#  define DEBUGP(fmt,args...)
#endif

enum conn_type {
  CLUSTER,  /* socket w/ listen() + bind()    */
  INCOMING, /* socket returned by accept()    */
  OUTGOING, /* socket w/ successful connect() */
  BACKEND,  /* backend machine definition     */
  DELETED,  /* marked for removal             */
  PENDING,  /* socket w/ connect() pending    */
};

/* our own string type */
struct String {
  char *s;
  uint64_t l, _l;
};

/* a connection */
struct conn {
  struct conn *next, *parent, *sibling;
  enum conn_type ct;
  int af, pf, sock;
  socklen_t sa_len;
  char *addr, *port;
  unsigned long long bytes, conns;
  time_t conn_time, error_time;
  struct String buffer;
  union {
    struct sockaddr_in sai4;
    struct sockaddr_in6 sai6;
    struct sockaddr_un sun;
  } sa;
};

int grace;

/* fatal error */
void nomem() {
  write(2, "no memory!\n", 11);
  exit(111);
}

int ultoa (const uint64_t l, char *buf) {
  static uint64_t save;
  static int len;
  static char *p;

  save = l;
  len = 0;
  p = buf;
  do {
   ++len;
   ++p;
   save /= 10;
  } while (save);
  save = l;
  do {
    --p;
    *p = '0' + save % 10;
    save /= 10;
  } while (save);
  return len;
}

/* String functions */
struct String *str_realloc (struct String *s, uint64_t l) {
  l = ((l + PAGE_SIZE - 1) / PAGE_SIZE) * PAGE_SIZE;
  if (l == s->_l) return s;
  s->s = realloc(s->s, l);
  if (!s->s) nomem();
  s->_l = l;
  return s;
}

struct String *str_cat_n (struct String *s, const char *c, const uint64_t l) {
  s = str_realloc(s, s->l + l + 1);
  memcpy(s->s + s->l, c, l);
  s->l += l;
  *(s->s + s->l) = '\0';
  return s;
}

struct String *str_cat (struct String *s, const char *c) {
  return str_cat_n(s, c, strlen(c));
}

struct String *str_cat_ui (struct String *s, uint64_t ui) {
  s = str_realloc(s, s->l + 21);
  s->l += ultoa(ui, s->s + s->l);
  *(s->s + s->l) = '\0';
  return s;
}

void dt_helper (int now, int *old, char *buf) {
  if (now == *old) return;
  *old = now;
  if (now < 10) *buf++ = '0';
  ultoa(now, buf);
}

struct String *str_cat_datetime (struct String *s) {
  static struct tm *nowtm, oldtm;
  static time_t now, old = 0;
  static char tmp[30] = "[01/Jan/1970:00:00:00 +0000] ";
  static const char *mon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

  now = time(NULL);

  if (now != old) {
    nowtm = gmtime(&now);
    dt_helper(nowtm->tm_mday, &oldtm.tm_mday, tmp + 1);
    if (nowtm->tm_mon != oldtm.tm_mon) {
      oldtm.tm_mon = nowtm->tm_mon;
      memcpy(tmp + 4, mon[oldtm.tm_mon], 3);
    }
    dt_helper(nowtm->tm_year + 1900, &oldtm.tm_year, tmp + 8);
    dt_helper(nowtm->tm_hour, &oldtm.tm_hour, tmp + 13);
    dt_helper(nowtm->tm_min, &oldtm.tm_min, tmp + 16);
    dt_helper(nowtm->tm_sec, &oldtm.tm_sec, tmp + 19);
  }

  return str_cat_n(s, tmp, 29);
}

/* non-fatal warnings */
void warn (char *s1, char *s2) {
  static struct String msg = NULLSTRING;

  msg.l = 0;
  str_cat_datetime(&msg);
  str_cat(&msg, s1);
  if (s2) {
    str_cat_n(&msg, ": ", 2);
    str_cat(&msg, s2);
  }
  str_cat_n(&msg, "\n", 1);
  write(2, msg.s, msg.l);
}

/* fatal warnings */
void die (char *s1, char *s2) {
  warn(s1, s2);
  exit(1);
}

struct String *str_inet_ntop (struct String *s, int af, void *sockaddr) {
  s = str_realloc(s, s->l + 512);
  switch (af) {
    case AF_INET:
      inet_ntop(af, (void*) &((struct sockaddr_in*) sockaddr)->sin_addr,
                s->s + s->l, 512);
      break;
    case AF_INET6:
      inet_ntop(af, (void*) &((struct sockaddr_in6*) sockaddr)->sin6_addr,
                s->s + s->l, 512);
      break;
    case AF_UNIX:
      str_cat(s, ((struct sockaddr_un*) sockaddr)->sun_path);
      return s;
    default:
      die("unknown protocol", NULL);
  }
  s->l += strlen(s->s + s->l);
  return s;
}

/* connection pool */
static int free_conns = 0;
static struct conn *free_list = NULL;

struct conn *new_conn (struct conn **list) {
  static struct conn *newc;

  if (free_conns == 0) {
    DEBUGP("%s", "filling conn pool");
    for (; free_conns < CONN_POOL_SIZE; ++free_conns) {
      newc = malloc(sizeof(struct conn));
      if (!newc) nomem();
      newc->buffer = NULLSTRING;
      newc->next = free_list;
      free_list = newc;
    }
  }
  else newc = free_list;
  free_list = newc->next;
  --free_conns;
  newc->conns = 0;
  newc->bytes = 0;
  newc->conn_time = time(NULL);
  newc->error_time = -1;
  newc->next = *list;
  *list = newc;
  return newc;
}

void clear_list (struct conn **list) {
  static struct conn *tmpc, *prevc, *nextc;

  for (prevc = NULL, tmpc = *list; tmpc; tmpc = nextc) {
    nextc = tmpc->next;
    if (tmpc->ct == DELETED) {
      if (prevc) prevc->next = nextc;
      else *list = nextc;
      tmpc->next = free_list;
      ++free_conns;
    }
    else
      prevc = tmpc;
  }
}

void gc () {
  static struct conn *delc;

  DEBUGP("free conns: %i", free_conns);
  if (free_conns >= CONN_POOL_SIZE) {
    DEBUGP("%s", "vacuuming conn pool");
    for (; free_conns > 0; --free_conns) {
      delc = free_list;
      free_list = delc->next;
      if (delc->buffer.s) free(delc->buffer.s);
      free(delc);
    }
  }
}

/* close bi-directional connection and make httpd compliant log entry */
void real_shutdown_conn (struct conn *delc, int epfd) {
  static int err;
  LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_DEL, delc->sock, NULL));
  if (err < 0) die("epoll_ctl()", strerror(errno));
  LOOPED_SYSCALL(close(delc->sock));
  delc->ct = DELETED;
}

#define shutdown_conn(delc,epfd,why,synproof) _shutdown_conn(delc,epfd,why,synproof,1)
void _shutdown_conn (struct conn *delc, int epfd, char *why, int synproof,
                     int del_incoming) {
  static struct conn *incom, *outgo;
  static struct String s = NULLSTRING;

  s.l = 0;
  if (delc->ct == INCOMING) {
    incom = delc;
    outgo = delc->sibling;
  }
  else {
    incom = delc->sibling;
    outgo = delc;
  }
  str_inet_ntop(&s, incom->af, (void*) &incom->sa);
  str_cat_n(&s, " - - ", 5);
  str_cat_datetime(&s);
  str_cat_n(&s, "\"", 1);
  if (synproof) str_cat_n(&s, "POST ", 5);
  else str_cat_n(&s, "GET ", 4);
  if (outgo) str_inet_ntop(&s, outgo->parent->af, (void*) &outgo->parent->sa);
  else str_cat_n(&s, "-", 1);
  str_cat_n(&s, "\" ", 2);
  str_cat(&s, why);
  str_cat_n(&s, " ", 1);
  str_cat_ui(&s, incom->bytes + ((outgo)?(outgo->bytes):(0)));
  str_cat_n(&s, "\n", 1);
  /*
   * remotehost rfc931(ident) authuser [date] "request" status bytes referer user-agent
   */
  write(1, s.s, s.l);
  if (del_incoming) real_shutdown_conn(incom, epfd);
  if (outgo) real_shutdown_conn(outgo, epfd);
}

/* make connection to backend as given by schedule function */
int new_backend_conn (struct conn **backends, struct conn **outgoings,
                      struct conn *sibling, int epfd,
                      struct conn* (*get_backend) (struct conn**),
                      int synproof) {
  static struct conn *newc, *parent;
  static struct epoll_event epev;
  static int err, flag;

  parent = get_backend(backends);
  if (!parent) {
    warn("no more backends!", NULL);
    shutdown_conn(sibling, epfd, "503", synproof);
    return -1;
  }
  newc = new_conn(outgoings);
  newc->parent = parent;
  newc->sibling = sibling;
  sibling->sibling = newc;
  newc->ct = PENDING;
  newc->af = parent->af;
  newc->pf = parent->pf;
  newc->sa_len = parent->sa_len;
  LOOPED_SYSCALL(newc->sock = socket(newc->pf, SOCK_STREAM, 0));
  if (newc->sock < 0) die("socket()", strerror(errno));
  LOOPED_SYSCALL(err = fcntl(newc->sock, F_GETFL, 0));
  if (err < 0) die("fcntl()", strerror(errno));
  flag = err | O_NONBLOCK;
  LOOPED_SYSCALL(err = fcntl(newc->sock, F_SETFL, flag));
  if (err < 0) die("fcntl()", strerror(errno));
  LOOPED_SYSCALL(flag = connect(newc->sock, (struct sockaddr*) &parent->sa,
                                parent->sa_len));
flag = errno;
  epev.events = EPOLLIN | EPOLLPRI | EPOLLOUT;
  epev.data.ptr = (void*) newc;
  LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_ADD, newc->sock, &epev));
  if (err < 0) die("epoll_ctl()", strerror(errno));
  if ((flag != 0) && (flag != EINPROGRESS)) {
    warn("connect()", strerror(flag));
    shutdown_conn(newc, epfd, "503", synproof);
  }
  else ++parent->conns;
  return 0;
}

/* returns NULL if connection is still in error condition */
struct conn *valid_backend (struct conn *con)
{
  return con->error_time + grace <= time(NULL) ? con : NULL;
}

/* schedulers */
struct conn *get_backend_rr (struct conn **backends) {
  static struct conn *ret = NULL;
  struct conn *start;

  if (ret) ret = ret->next;
  if (!ret) ret = *backends;
  start = ret;
  while (!valid_backend(ret)) {
    ret = ret->next;
    if (!ret) ret = *backends;
    if (ret == start) return NULL;
  }
  return ret;
}

struct conn *get_backend_minconns (struct conn **backends) {
  static struct conn *ret, *tmp;
  ret = *backends;
  for (tmp = ret; tmp != NULL; tmp = tmp->next)
    if (valid_backend(tmp) && (tmp->conns < ret->conns)) ret = tmp;
  return valid_backend(ret);
}

struct conn *get_backend_minbytes (struct conn **backends) {
  static struct conn *ret, *tmp;
  ret = *backends;
  for (tmp = ret; tmp != NULL; tmp = tmp->next)
    if (valid_backend(tmp) && (tmp->bytes < ret->bytes)) ret = tmp;
  return valid_backend(ret);
}

struct conn *get_backend_mintimes (struct conn **backends) {
  static struct conn *ret, *tmp;
  ret = *backends;
  for (tmp = ret; tmp != NULL; tmp = tmp->next)
    if (valid_backend(tmp) && (tmp->conn_time < ret->conn_time)) ret = tmp;
  return valid_backend(ret);
}

unsigned long long algo_weighted (struct conn *con) {
  return con->conns + con->bytes / 1024 + con->conn_time;
}

struct conn *get_backend_weighted (struct conn **backends) {
  static struct conn *ret, *tmp;
  static unsigned long long old, new;
  ret = *backends;
  old = algo_weighted(ret);
  for (tmp = ret; tmp != NULL; tmp = tmp->next) {
    if (valid_backend(ret)) {
      new = algo_weighted(tmp);
      if (new < old) {
        ret = tmp;
        old = new;
      }
    }
  }
  return valid_backend(ret);
}

/* wrapper for sigaction */
void signal_set (int sig, void (*handler)(int), int check_err)
{
  static struct sigaction sigact;
  static int err;

  memset(&sigact, 0, sizeof(struct sigaction));
  sigact.sa_handler = handler;
  LOOPED_SYSCALL(err = sigaction(sig, &sigact, NULL));
  if (!check_err) return;
  if (err < 0) die("sigaction()", strerror(errno));
}

/* signal handling */
int reopen_logfiles = 0, exit_nicely = 0;
void sighup_handler (int i) {
  i = i;
  reopen_logfiles = 1;
}

void sigterm_handler (int i) {
  i = i;
  exit_nicely = 1;
}

/* exchange underlying file for descriptor */
void reopen (char *fn, int fd) {
  static int nfd, err;
  LOOPED_SYSCALL(nfd = open(fn, O_CREAT | O_APPEND | O_WRONLY,
                            S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
  if (nfd < 0) {
    warn(fn, strerror(errno));
    return;
  }
  LOOPED_SYSCALL(err = dup2(nfd, fd));
  if (err < 0) die("dup2()", strerror(errno)); /* maybe not visible */
  LOOPED_SYSCALL(close(nfd));
}

/* show some help */
void usage ()
{
  static const char *help =
"fproxy V" VERSION " - non-forking tcp proxy using the epoll() framework\n\n"
"fproxy [-o<logfile>] [-e<errorlog>] [-p<pidfile>] [-f] [-y<# bytes>]\n"
"       [-t<timeout>] [-g<grace>] [-s0] [-s1] [-s2] [-s3] [-s4]\n"
"       <cluster[:port]> [ ... ] -- <backend[:port]> [ ... ]\n\n"
"fproxy -h\n\n"
"  -h               this help :-)\n"
"  -o<logfile>      use <logfile> for logging instead of /var/log/fproxy.log\n"
"  -e<errorlog>     use <errorlog> for error messages\n"
"                   instead of /var/log/fproxy.err\n"
"  -p<pidfile>      write process id into <pidfile>,\n"
"                   not into /var/run/fproxy.pid\n"
"  -f               do not fork and detach from controlling terminal;\n"
"                   do not use logfile, errorlog and pidfile\n"
"  -y<# bytes>      open new connection to backend servers only if the client\n"
"                   wrote <# bytes>, eg. use -y6 for preventing a SYN attack\n"
"                   against a webserver (nb. strlen(\"GET /\\n\") = 6);\n"
"                   -y0 disables this feature, which is also the default\n"
"  -t<timeout>      drop idle client connections after <timeout> seconds;\n"
"                   -t-1 disables timeout handling, which is also the default\n"
"  -g<grace>        reenables a failed backend after <grace> seconds;\n"
"                   defaults to 60 seconds\n"
"  -s0              select backends in a round robin fashion\n"
"  -s1              select backend with least connections handled so far\n"
"  -s2              select backend with least bytes (in + out) processed so far\n"
"  -s3              select backend with least connection time consumed so far\n"
"  -s4              (default) select backend in a weighted fashion, using number\n"
"                   of connections, processed bytes and connection time to\n"
"                   calculate to workload of each backend\n"
"  <cluster[:port]> path to Unix socket or IPv4 or IPv6 address(es) w/ optional\n"
"                   tcp port on which this machine accepts connections;\n"
"                   port defaults to 80\n"
"  <backend[:port]> path to Unix socket or IPv4 or IPv6 address(es) w/ optional\n"
"                   tcp port to which client connections be proxied;\n"
"                   port defaults to 80\n\n"
" send SIGHUP  to reopen logfile and errorlog\n"
" send SIGTERM to exit nicely (dont accept any new client connections, wait\n"
"              for active connections to be finished normally)\n"
" send SIGKILL to exit immediatelly\n\n";
  write(1, help, strlen(help));
}

/* main routine */
int main (int argc, char **argv) {
  enum conn_type ct;
  struct conn *newc, *tmpc,
    *clusters = NULL, *incomings = NULL,
    *outgoings = NULL, *backends = NULL;
  struct in_addr ia4;
  struct in6_addr ia6;
  int err, epfd, epevs, timeout = -1;
  unsigned int synproof = 0, dontfork = 0;
  ssize_t len;
  struct epoll_event epev[EPOLL_HINT];
  time_t now;
  char *logfile = LOGFILE, *errfile = ERRFILE, *pidfile = PIDFILE,
    buffer[PAGE_SIZE];
  struct conn* (*get_backend_func) (struct conn**) = get_backend_rr;

  epfd = epoll_create(EPOLL_HINT);
  if (epfd < 0) die("epoll_create()", strerror(errno));

  grace = 60;

  /* evaluate arguments */
  ct = BACKEND;
  for (--argc; argc > 0; --argc) {
    if (!strcmp(argv[argc], "--")) {
      ct = CLUSTER;
      continue;
    }
    if (!strncmp(argv[argc], "-o", 2)) {
      logfile = argv[argc] + 2;
      continue;
    }
    if (!strncmp(argv[argc], "-e", 2)) {
      errfile = argv[argc] + 2;
      continue;
    }
    if (!strncmp(argv[argc], "-p", 2)) {
      pidfile = argv[argc] + 2;
      continue;
    }
    if (!strcmp(argv[argc], "-f")) {
      dontfork = 1;
      continue;
    }
    if (!strncmp(argv[argc], "-y", 2)) {
      synproof = atoi(argv[argc] + 2);
      DEBUGP("synproof:%i", synproof);
      continue;
    }
    if (!strncmp(argv[argc], "-t", 2)) {
      timeout = atoi(argv[argc] + 2);
      DEBUGP("timeout:%i", timeout);
      continue;
    }
    if (!strncmp(argv[argc], "-g", 2)) {
      grace = atoi(argv[argc] + 2);
      DEBUGP("grace:%i", grace);
      continue;
    }
    if (!strcmp(argv[argc], "-h")) {
      usage();
      return 0;
    }
    if (!strcmp(argv[argc], "-s0")) {
      get_backend_func = get_backend_rr;
      continue;
    }
    if (!strcmp(argv[argc], "-s1")) {
      get_backend_func = get_backend_minconns;
      continue;
    }
    if (!strcmp(argv[argc], "-s2")) {
      get_backend_func = get_backend_minbytes;
      continue;
    }
    if (!strcmp(argv[argc], "-s3")) {
      get_backend_func = get_backend_mintimes;
      continue;
    }
    if (!strcmp(argv[argc], "-s4")) {
      get_backend_func = get_backend_weighted;
      continue;
    }

    newc = new_conn((ct == BACKEND)?(&backends):(&clusters));
    newc->ct = ct;
    newc->addr = argv[argc];
    newc->port = strrchr(newc->addr, ':');
    if (newc->port) *newc->port++ = '\0';
    else newc->port = "80";

    if (argv[argc][0] == '/') {
      newc->af = AF_UNIX;
      newc->pf = PF_UNIX;
      newc->sa_len = sizeof(struct sockaddr_un);
      memcpy(&newc->sa.sun.sun_path, argv[argc], strlen(argv[argc]) + 1);
      newc->sa.sun.sun_family = AF_UNIX;
    }
    else if (inet_pton(AF_INET, argv[argc], (void*) &ia4)) {
      newc->af = AF_INET;
      newc->pf = PF_INET;
      newc->sa_len = sizeof(struct sockaddr_in);
      memcpy(&newc->sa.sai4.sin_addr, &ia4, sizeof(struct in_addr));
      newc->sa.sai4.sin_port = htons(atoi(newc->port));
      newc->sa.sai4.sin_family = AF_INET;
    }
    else if (inet_pton(AF_INET6, argv[argc], (void*) &ia6)) {
      newc->af = AF_INET6;
      newc->pf = PF_INET6;
      newc->sa_len = sizeof(struct sockaddr_in6);
      memcpy(&newc->sa.sai6.sin6_addr, &ia6, sizeof(struct in6_addr));
      newc->sa.sai6.sin6_port = htons(atoi(newc->port));
      newc->sa.sai6.sin6_family = AF_INET6;
    }
    else die("invalid address", argv[argc]);

    if (ct != CLUSTER) continue;

    for (tmpc = backends; tmpc; tmpc = tmpc->next)
      if (!memcmp((void*) &tmpc->sa, (void*) &newc->sa, newc->sa_len) &&
          (tmpc->sa_len == newc->sa_len))
        die("backend <-> cluster loop detected", argv[argc]);

    newc->sock = socket(newc->pf, SOCK_STREAM, 0);
    if (newc->sock < 0) die("socket()", strerror(errno));
    if (bind(newc->sock, (struct sockaddr*) &newc->sa, newc->sa_len) < 0)
      die("bind()", strerror(errno));
    if (listen(newc->sock, LISTEN_BACKLOG) < 0)
      die("listen()", strerror(errno));
    epev[0].events = EPOLLIN | EPOLLPRI;
    epev[0].data.ptr = (void*) newc;
    if (epoll_ctl(epfd, EPOLL_CTL_ADD, newc->sock, &epev[0]) < 0)
      die("epoll_ctl()", strerror(errno));
  }

  if (clusters == NULL) die("no clusters!", NULL);
  if (backends == NULL) die("no backends!", NULL);

  for (err = 1; err <= SIGRTMAX; ++err) signal_set(err, SIG_IGN, 0);
  signal_set(SIGHUP, sighup_handler, 1);
  signal_set(SIGTERM, sigterm_handler, 1);

  if (!dontfork) {
    /* daemonize */
    pid_t pid;
    reopen(logfile, 1);
    reopen(errfile, 2);
    pid = fork();
    if (pid < 0) die("fork()", strerror(errno));
    if (pid != 0) return 0;
    umask(0022);
    if (chdir("/") < 0) die("chdir()", strerror(errno));
    close(0);
    if (setsid() < 0) die("setsid()", strerror(errno));
    LOOPED_SYSCALL(err = open(pidfile, O_CREAT | O_TRUNC | O_WRONLY,
                              S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
    if (err < 0) die(pidfile, strerror(errno));
    epevs = ultoa(getpid(), buffer);
    len = write(err, buffer, epevs);
    if (len < 0) die("write()", strerror(errno));
    if (len != epevs) die(pidfile, "cannot write all data!?");
    close(err);
  }

  /* enter loop */
  warn("fproxy", "running");
  while ((exit_nicely != 2) || incomings) {
    epevs = epoll_wait(epfd, epev, EPOLL_HINT, EPOLL_TIMEOUT);
    if ((epevs < 0) && (errno != EINTR)) die("epoll_wait()", strerror(errno));
    if (epevs > EPOLL_HINT)
      die("epoll_wait()", "returned more than EPOLL_HINT events!?!");
    now = time(NULL);
    for (--epevs; epevs >= 0; --epevs) {
      tmpc = (struct conn*) epev[epevs].data.ptr;
      switch (tmpc->ct) {
        case CLUSTER:
          newc = new_conn(&incomings);
          newc->parent = tmpc;
          newc->sibling = NULL;
          newc->ct = INCOMING;
          newc->af = tmpc->af;
          newc->pf = tmpc->pf;
          newc->sa_len = tmpc->sa_len;
          LOOPED_SYSCALL(newc->sock = accept(tmpc->sock,
                                             (struct sockaddr*) &newc->sa,
                                             &newc->sa_len));
          if (newc->sock < 0) {
            warn("accept()", strerror(errno));
            warn(tmpc->addr, "shutting down cluster address due to accept() "
                             "error");
            LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_DEL, tmpc->sock,
                                           NULL));
            if (err < 0) die("epoll_ctl()", strerror(errno));
            LOOPED_SYSCALL(err = close(tmpc->sock));
            if (err < 0) warn("close()", strerror(errno));
            tmpc->ct = DELETED;
            newc->ct = DELETED;
            break;
          }
          ++tmpc->conns;
          epev[epevs].events = EPOLLIN | EPOLLPRI;
          epev[epevs].data.ptr = (void*) newc;
          LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_ADD, newc->sock,
                                         &epev[epevs]));
          if (err < 0) die("epoll_ctl()", strerror(errno));
          if (!synproof)
            new_backend_conn(&backends, &outgoings, newc, epfd,
                             get_backend_func, synproof);
          break;
        case INCOMING:
        case OUTGOING:
          tmpc->parent->conn_time += now - tmpc->conn_time;
          tmpc->conn_time = now;
          if ((epev[epevs].events & EPOLLHUP) ||
              (epev[epevs].events & EPOLLERR)) {
            shutdown_conn(tmpc, epfd, "503", synproof);
            break;
          }
          if ((epev[epevs].events & EPOLLIN) ||
              (epev[epevs].events & EPOLLPRI)) {
            str_realloc(&tmpc->buffer, tmpc->buffer.l + PAGE_SIZE);
            LOOPED_SYSCALL(len = recv(tmpc->sock,
                                      tmpc->buffer.s + tmpc->buffer.l,
                                      PAGE_SIZE, MSG_DONTWAIT));
            if (len < 0) {
              warn("recv()", strerror(errno));
              shutdown_conn(tmpc, epfd, "500", synproof);
              break;
            }
            if (len == 0) {
              shutdown_conn(tmpc, epfd, "200", synproof);
              break;
            }
            tmpc->buffer.l += len;
            tmpc->bytes += len;
            tmpc->parent->bytes += len;
            if (!tmpc->sibling) {
              if (tmpc->buffer.l < synproof) break;
              if (new_backend_conn(&backends, &outgoings, tmpc, epfd,
                                   get_backend_func, synproof) < 0) break;
            }
            if (tmpc->sibling->ct == PENDING) break;
            LOOPED_SYSCALL(len = send(tmpc->sibling->sock, tmpc->buffer.s,
                                      tmpc->buffer.l,
                                      MSG_DONTWAIT | MSG_NOSIGNAL));
            if (len < 0) {
              if (errno == EAGAIN) break;
              if (errno == EWOULDBLOCK) break;
              warn("send()", strerror(errno));
              shutdown_conn(tmpc, epfd, "500", synproof);
              break;
            }
            tmpc->buffer.l -= len;
            memmove(tmpc->buffer.s, tmpc->buffer.s + len, tmpc->buffer.l);
          }
          break;
        case PENDING:
          if ((epev[epevs].events & EPOLLHUP) ||
              (epev[epevs].events & EPOLLERR)) {
            warn(tmpc->parent->addr, "disabling backend address due to "
                                     "connect() error");
            tmpc->parent->error_time = now;
            _shutdown_conn(tmpc, epfd, "503", synproof, 0);
            new_backend_conn(&backends, &outgoings, tmpc->sibling, epfd,
                             get_backend_func, synproof);
            break;
          }
          tmpc->ct = OUTGOING;
          epev[epevs].events = EPOLLIN | EPOLLPRI;
          LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_MOD, tmpc->sock,
                                         &epev[epevs]));
          if (err < 0) die("epoll_ctl()", strerror(errno));
          LOOPED_SYSCALL(len = send(tmpc->sock, tmpc->sibling->buffer.s,
                                    tmpc->sibling->buffer.l,
                                    MSG_DONTWAIT | MSG_NOSIGNAL));
          if (len < 0) {
            if (errno == EAGAIN) break;
            if (errno == EWOULDBLOCK) break;
            warn("send()", strerror(errno));
            shutdown_conn(tmpc, epfd, "500", synproof);
            break;
          }
          tmpc->sibling->buffer.l -= len;
          memmove(tmpc->sibling->buffer.s, tmpc->sibling->buffer.s + len,
                  tmpc->sibling->buffer.l);
          break;
        case DELETED:
          DEBUGP("%s", "encountered DELETED conn");
          break;
        default:
          ultoa(tmpc->ct, buffer);
          die("epoll event on unhandeled connection type", buffer);
      } /* switch */
    } /* for */

    /* remove timed out connections */
    if (timeout != -1) for (tmpc = incomings; tmpc; tmpc = tmpc->next)
      if (now - tmpc->conn_time > timeout)
        shutdown_conn(tmpc, epfd, "502", synproof);

    /* sighup? */
    if (reopen_logfiles == 1) {
      if (dontfork)
        warn("SIGHUP received", "cannot reopen logfiles in non-server mode");
      else {
        reopen(logfile, 1);
        reopen(errfile, 2);
      }
      reopen_logfiles = 0;
    }

    /* sigkill? */
    if (exit_nicely == 1) {
      unlink(pidfile);
      warn("SIGKILL received", "waiting for connections to finish");
      for (; clusters; clusters = clusters->next) {
        if (clusters->ct == DELETED) continue;
        LOOPED_SYSCALL(err = epoll_ctl(epfd, EPOLL_CTL_DEL,
                                       clusters->sock, NULL));
        if (err < 0) die("epoll_ctl()", strerror(errno));
        LOOPED_SYSCALL(close(clusters->sock));
        if (*clusters->addr == '/') LOOPED_SYSCALL(unlink(clusters->addr));
      }
      exit_nicely = 2;
    }

    /* cleanup */
    clear_list(&incomings);
    clear_list(&outgoings);
    gc();
  } /* while */

  warn("no incomings any more", "going down");
  return 0;
}
